/* -*-	Mode:C++; c-basic-offset:4; tab-width:4; indent-tabs-mode:t -*- */

/*
 * Copyright (c) 2014, Damien Magoni.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Damien Magoni.
 * 4. The name of Damien Magoni may not be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY DAMIEN MAGONI ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL DAMIEN MAGONI OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

// DMP for Partial Data Aggregation

#include "dmp.h"
#define TCL_OK 0

// global variable (local to this file)
demigod DG;

int hdr_dmp::offset_;

static class DMPHeaderClass : public PacketHeaderClass
{
public:
	DMPHeaderClass() : PacketHeaderClass("PacketHeader/DMP", sizeof(hdr_dmp))
	{
		bind_offset(&hdr_dmp::offset_);
	}
} class_dmphdr;

static class DmpClass : public TclClass
{
public:
	DmpClass() : TclClass("Agent/DMP") {}
	TclObject * create(int argc, const char * const * argv)
	{
		return (new DmpAgent());
	}
} class_dmp;

void DMPDataTimer::expire(Event * e)
{
	status_ = TIMER_IDLE;
	a_->data_timeout();
}

void DMPLocalAggTimer::expire(Event * e)
{
	status_ = TIMER_IDLE;
	a_->local_aggr_timeout();
}

void DMPGlobalAggTimer::expire(Event * e)
{
	status_ = TIMER_IDLE;
	a_->global_aggr_timeout();
}

void DMPBeaconTimer::expire(Event * e)
{
	status_ = TIMER_IDLE;
	a_->beacon_timeout();
}

DmpAgent::DmpAgent() : Agent(PT_DMP),
		beacon_timer(this), data_timer(this), global_agg_timer(this), local_agg_timer(this)
{
	bind("node_id_", &node_id_);
	bind("node_address_", &node_address_);
	bind("min_connect_", &min_connect_);
	bind("net_type_", &net_type_);
	bind("nodes_nb_", &nodes_nb_);
	bind("verbose_", &verbose_);
	bind("seed_", &seed_);
	bind("algorithm_", &algorithm_);
	bind("message_period_", &message_period_);
	bind("local_wait_period_", &local_wait_period_);
	bind("global_wait_period_", &global_wait_period_);
	bind("beacon_period_", &beacon_period_);
	bind("tx_max_", &tx_max_);
	bind("margin_", &margin_);
	bind("rx_thr_", &rx_thr_);
	bind("range_max_", &range_max_);

	sensor_type_ = SENSOR_CM;
	CM_nb_ = 0;
	cluster_id_ = 0;
	min_local_values = 0.2; // as a ratio
	min_global_values = 3;

	cv_local_threshold = 0.25;
	cv_global_threshold = 0.25;

	pkt_recv = 0;
	pkt_sent = 0;
	pkt_sent_fast = 0;
	spatial_closeness_ = 25; // m
	x = 0;
	y = 0;
	energy_used = 0;
	transmission_range = range_max_; // 85.46029; // necessary?
	transmission_power = tx_max_; // max tx power in dBm
	transmission_power_change_time = 0;
	dg = &DG;
	neighbors.clear();
	maxneighbors.clear();
	communicators.clear();
}

int DmpAgent::command(int argc, const char * const * argv)
{
	if ( argc == 2 )
	{
		if ( !strcmp(argv[1], "dmpstart") )
		{
			start();
			return (TCL_OK);
		}
		if ( !strcmp(argv[1], "dmpstat") )
		{
			stat();
			return (TCL_OK);
		}
	}
	return (Agent::command(argc, argv));
}

void DmpAgent::stat()
{
	dg->output();
}

void DmpAgent::start()
{
	if ( verbose_ ) printf("Starting node having address %i on a total of %i\n", node_address_, nodes_nb_);
	// set Node and Mac pointers
	Tcl& tcl = Tcl::instance();
	char inbuf[256];
	sprintf(inbuf,"%s set node_handle_", name());
	tcl.eval(inbuf);
	const char * res = tcl.result();
	char outbuf[256];
	strcpy(outbuf, res);
	//std::cout << "lookup for node named " << outbuf << "\n";
	node_pt = (Node *)tcl.lookup(outbuf); // rather use MobileNode? no, plus this seems useless

	sprintf(inbuf,"%s set mac_handle_", name());
	tcl.eval(inbuf);
	res = tcl.result();
	strcpy(outbuf, res);
	//std::cout << "lookup for mac named " << outbuf << "\n";
	mac_pt = (Mac *)tcl.lookup(tcl.result());
	mac_pt->dmp_agent_pt = this;
	mac_address_ = mac_pt->index_;

	set_x();
	set_y();
	dg->init(this);
	if ( node_id_ )
	{
		epoch = 0;
		data_timer.sched(message_period_ * dmp::rzto()); // in seconds
		if ( algorithm_ != aggregation_none )
		{
			if ( true )//algorithm_ == aggregation_random )
			{
				local_agg_timer.sched(local_wait_period_ * dmp::rzto()); // in seconds
				global_agg_timer.sched(global_wait_period_ * dmp::rzto()); // in seconds
				//global_agg_timer.sched(global_wait_period_); // in seconds
			}
			else // all nodes will aggregate at the same time: maybe that's not good
			{
				local_agg_timer.sched(local_wait_period_); // in seconds
				global_agg_timer.sched(global_wait_period_); // in seconds
			}
		}
	}
}

void DmpAgent::recv(Packet * p, Handler * h)
{
	if ( verbose_ ) printf("Receiving packet at %i\n", node_address_);
	hdr_dmp * hd = hdr_dmp::access(p);
	switch ( hd->dmp_subtype_ )
	{
		case DMP_BEACON :
		{
			if ( verbose_ ) printf("Beacon received by %i\n", node_address_);
			process_beacon(p);
		}
		break;
		case DMP_BEACON_REPLY :
		{
			if ( verbose_ ) printf("Reply received by %i\n", node_address_);
			process_reply(p);
		}
		break;
		case DMP_DATA :
		{
			if ( verbose_ ) printf("Data received by %i\n", node_address_);
			process_data(p);
		}
		break;
		default :
		{
			if ( verbose_ ) printf("Unknown packet subtype received by %i\n", node_address_);
			;
		}
		break;
	}
	Packet::free(p);
}

void DmpAgent::process_data(Packet * p)
{
	//if ( verbose_ ) std::cout << "Reached destination/sink having addr " << node_address_ << "\n";
	//hdr_cmn * hc = hdr_cmn::access(p);
	//hdr_ip * hi = hdr_ip::access(p);
	hdr_dmp * hd = hdr_dmp::access(p);
	e2e_delay.push_back(Scheduler::instance().clock() - hd->timestamp_);
}

int DmpAgent::hook(Packet *p) // from Mac up to Agent
{
	if ( verbose_ ) std::cout << "Hooking in node " << node_address_ << "\n";
	//hdr_cmn * hc = hdr_cmn::access(p);
	hdr_ip * hi = hdr_ip::access(p);
	hdr_dmp * hd = hdr_dmp::access(p);
	if ( hd->dmp_subtype_ != DMP_DATA )
	{
		std::cout << "unknown subtype, let it through\n";
		return 1;
	}
	if ( hi->dst_.addr_ == node_address_ )
	{
		return 1; // in dest/sink, let it to recv
	}
	pkt_recv++;
	if ( algorithm_ != aggregation_none ) // intermediate node
	{
		aggregate_data(p);
		Packet::free(p);
		return 0; // consume
	}
	else // algorithm 1: no aggreg
	{
		pkt_sent++;
		dg->tot_pkts_sent++;
		return 1; // passthrough
	}
}

void DmpAgent::inject(Packet * p) // from Agent down to Mac, we inject only new packets
{
	if ( !p )
	{
		//p = allocpkt();
		std::cout << "error, can't inject null packet\n";
		return;
	}
	hdr_cmn * hc = hdr_cmn::access(p);
	hc->ptype_ = PT_DMP;
	//hc->direction_ = hdr_cmn::DOWN; // outgoing packet
	hc->size_ = 64;
	hc->ts_ = Scheduler::instance().clock();
	hc->addr_type_ = NS_AF_INET;
	hc->last_hop_ = node_address_;
	hc->prev_hop_ = node_address_;
	//nsaddr_t next_hop_;	// next hop for this packet, should be set by the RT and the LL

	hdr_ip * hi = hdr_ip::access(p);
	hi->src_.addr_ = node_address_;
	hi->dst_.addr_ = 0; // sink address
	hi->src_.port_ = 10;
	hi->dst_.port_ = 10;
	hi->ttl_ = 128;

	send(p, (Handler *)0); // sent by Agent base class to target_
	//mac_pt->recv(p, (Handler *)0); // bypass RT and LL which is bad
}

void DmpAgent::data_timeout()
{
	// create genuine non aggr packet
	if ( verbose_ ) std::cout << "Injecting new packet in node " << node_address_ << "\n";
	Packet * p = allocpkt();
	hdr_dmp * hd = hdr_dmp::access(p);
	hd->dmp_subtype_ = DMP_DATA;
	hd->priority_level_ = 0;
	hd->content_type_ = 0;
	hd->content_value_ = dmp::random(15, 25);
	hd->timestamp_ = Scheduler::instance().clock();
	hd->aggregation_ = 0;
	hd->sensor_type_ = SENSOR_CM;
	hd->cluster_id_ = cluster_id_;
	hd->x_ = x;
	hd->y_ = y;
	inject(p);
	dg->gen_pkts_sent++;
	dg->tot_pkts_sent++;
	epoch++; // true only if wait  = fixed message period
	data_timer.resched(message_period_);
}

void DmpAgent::local_aggr_timeout()
{
	int i;
	for ( i = 0; i < local_values.size(); i++ )
	{
		//if ( !local_values[i].partial ) // let's do it in any case, this doubles the nb of messages
		{
			std::vector<double> vv = dmp::extract_vec(local_values[i].values, false);
			std::vector<double> kv = dmp::extract_vec(local_values[i].values, true);
			if ( vv.size() > 0 )
			{
				Packet * p = allocpkt();
				hdr_dmp * hd = hdr_dmp::access(p);
				hd->dmp_subtype_ = DMP_DATA;
				hd->priority_level_ = 0;
				hd->content_type_ = 0;
				hd->content_value_ = dmp::sample_mean(vv);
				hd->timestamp_ = dmp::sample_min(kv);
				hd->x_ = x;
				hd->y_ = y;
				hd->aggregation_ = 1;
				pkt_sent++;
				dg->tot_pkts_sent++;
				inject(p);
			}
		}
	}
	local_values.clear();
	if ( algorithm_ == aggregation_random )
	{
		double complement = 0;//local_wait_period_ - (dmp::round(Scheduler::instance().clock()) % dmp::round(local_wait_period_)); // in whole secs (due to %)
		local_agg_timer.resched(complement + local_wait_period_ * dmp::rzto());
	}
	else
		local_agg_timer.resched(local_wait_period_);
}

void DmpAgent::global_aggr_timeout()
{
	int i;
	for ( i = 0; i < global_values.size(); i++ )
	{
		//if ( !global_values[i].partial ) // see above
		{
			std::vector<double> vv = dmp::extract_vec(global_values[i].values, false);
			std::vector<double> kv = dmp::extract_vec(global_values[i].values, true);
			if ( vv.size() > 0 )
			{
				Packet * p = allocpkt();
				hdr_dmp * hd = hdr_dmp::access(p);
				hd->dmp_subtype_ = DMP_DATA;
				hd->priority_level_ = 0;
				hd->content_type_ = 0;
				hd->content_value_ = dmp::sample_mean(vv);
				hd->timestamp_ = dmp::sample_min(kv);
				hd->x_ = global_values[i].x;
				hd->y_ = global_values[i].y;
				hd->aggregation_ = 1;
				pkt_sent++;
				dg->tot_pkts_sent++;
				inject(p);
			}
		}
	}
	global_values.clear();
	if ( algorithm_ == aggregation_random )
	{
		double complement = 0;//global_wait_period_ - (dmp::round(Scheduler::instance().clock()) % dmp::round(global_wait_period_)); // in whole secs (due to %)
		global_agg_timer.resched(complement + global_wait_period_ * dmp::rzto());
	}
	else
		global_agg_timer.resched(global_wait_period_);
}

void DmpAgent::aggregate_data(Packet * p)
{
	if ( verbose_ ) std::cout << "Aggregating in " << node_address_ << "\n";
	hdr_cmn * hc = hdr_cmn::access(p);
	hdr_ip * hi = hdr_ip::access(p);
	bool is_local = false;
	if ( hc->prev_hop_ == hi->src_.addr_ ) // we are in CH
	{
		if ( Scheduler::instance().clock() <= message_period_ ) // we are in 1st period
			CM_nb_++;
		is_local = true;
	}
	store_data(is_local, p);
	if ( algorithm_ == aggregation_partial )
	{
		evaluate_partial_aggr(is_local, p);
	}
}

void DmpAgent::store_data(bool local, Packet * p)
{
	hdr_dmp * hd = hdr_dmp::access(p);
	entry e;
	e.priority_level = hd->priority_level_;
	e.content_type = hd->content_type_;
	e.partial = false;
	e.values.insert(std::pair<double, double>(hd->timestamp_, hd->content_value_));
	if ( local )
	{
		e.x = x; // x of CH
		e.y = y; // y of CH
		int i;
		for ( i = 0; i < local_values.size(); i++ )
		{
			if ( local_values[i].priority_level == e.priority_level && local_values[i].content_type == e.content_type )
			{
				local_values[i].values[hd->timestamp_] = hd->content_value_; // won't override existing entry
				break;
			}
		}
		if ( i == local_values.size() )
		{
			local_values.push_back(e);
		}
	}
	else
	{
		e.x = hd->x_;
		e.y = hd->y_;
		int i;
		for ( i = 0; i < global_values.size(); i++ )
		{
			if ( global_values[i].priority_level == e.priority_level && global_values[i].content_type == e.content_type )
			{
				double d = dmp::dist(global_values[i].x, global_values[i].y, e.x, e.y);
				if ( d <= spatial_closeness_ )
				{
					global_values[i].values[hd->timestamp_] = hd->content_value_; // won't override existing entry
					break;
				}
			}
		}
		if ( i == global_values.size() )
		{
			global_values.push_back(e);
		}
	}
}

void DmpAgent::evaluate_partial_aggr(bool local, Packet * p)
{
	hdr_dmp * hd = hdr_dmp::access(p);
	if ( local )
	{
		int i = 0;
		for ( i = 0; i < local_values.size(); i++ )
		{
			if ( local_values[i].priority_level == hd->priority_level_ && local_values[i].content_type == hd->content_type_ )
			{
				std::vector<double> vv = dmp::extract_vec(local_values[i].values, false);
				std::vector<double> kv = dmp::extract_vec(local_values[i].values, true);
				// test 1, look at the data
				//if ( local_values[i].values.size() >= min_local_values
				//	&& dmp::sample_standard_deviation(vv)/dmp::sample_mean(vv) <= cv_local_threshold )
				// test 2, look at the nb of replies from CMs
				if ( (CM_nb_ > 0) )
				{
					if ( static_cast<double>(local_values[i].values.size()) / CM_nb_ >= min_local_values )
					{
						local_values[i].partial = true;
						Packet * np = allocpkt();
						hdr_dmp * nhd = hdr_dmp::access(np);
						nhd->dmp_subtype_ = DMP_DATA;
						nhd->priority_level_ = 0;
						nhd->content_type_ = 0;
						nhd->content_value_ = dmp::sample_mean(vv);
						nhd->timestamp_ = dmp::sample_min(kv);
						nhd->x_ = local_values[i].x;
						nhd->y_ = local_values[i].y;
						nhd->aggregation_ = 1;
						pkt_sent++;
						pkt_sent_fast++;
						dg->tot_pkts_sent++;
						inject(np);
						local_values.erase(local_values.begin() + i); // otherwise delays are false!! -> v4
					}
				}
				break;
			}
		}
	}
	else
	{
		int i = 0;
		for ( i = 0; i < global_values.size(); i++ )
		{
			if ( global_values[i].priority_level == hd->priority_level_ && global_values[i].content_type == hd->content_type_ )
			{
				double d_n_p = dmp::dist(x, y, hd->x_, hd->y_); // distance from node to source CH
				double d_e_p = dmp::dist(global_values[i].x, global_values[i].y, hd->x_, hd->y_); // distance from entry to source CH
				// test 3 -> introduced in v5
				if ( d_e_p == 0 and d_n_p > spatial_closeness_ ) // 1st/new entry of this kind/class, far away no spatial aggr, send it: why?? XXX
				{
					if ( global_values[i].values.size() != 1 )
						std::cout << "error, not 1st entry\n";
					if ( true )
					{
						global_values[i].partial = true;
						Packet * np = allocpkt();
						hdr_dmp * nhd = hdr_dmp::access(np);
						nhd->dmp_subtype_ = DMP_DATA;
						nhd->priority_level_ = 0;
						nhd->content_type_ = 0;
						nhd->content_value_ = hd->content_value_;//dmp::sample_mean(vv);
						nhd->timestamp_ = hd->timestamp_;//dmp::sample_min(kv);
						nhd->x_ = hd->x_;//global_values[i].x;
						nhd->y_ = hd->y_;//global_values[i].y;
						nhd->aggregation_ = 1;
						pkt_sent++;
						pkt_sent_fast++;
						dg->tot_pkts_sent++;
						inject(np);
						global_values.erase(global_values.begin() + i); // correct? yes -> v4
					}
					break;
				}
				else if ( d_e_p <= spatial_closeness_ ) // the message has been aggr in store_data
				{
					std::vector<double> vv = dmp::extract_vec(global_values[i].values, false);
					std::vector<double> kv = dmp::extract_vec(global_values[i].values, true);
					// test 1
					//if ( global_values[i].values.size() >= min_global_values
					//	&& dmp::sample_standard_deviation(vv)/dmp::sample_mean(vv) <= cv_global_threshold )
					// test 2
					if ( static_cast<double>(global_values[i].values.size()) >= min_global_values ) // condition for partial send
					{
						global_values[i].partial = true;
						Packet * np = allocpkt();
						hdr_dmp * nhd = hdr_dmp::access(np);
						nhd->dmp_subtype_ = DMP_DATA;
						nhd->priority_level_ = 0;
						nhd->content_type_ = 0;
						nhd->content_value_ = dmp::sample_mean(vv);
						nhd->timestamp_ = dmp::sample_min(kv);
						nhd->x_ = global_values[i].x;
						nhd->y_ = global_values[i].y;
						nhd->aggregation_ = 1;
						pkt_sent++;
						pkt_sent_fast++;
						dg->tot_pkts_sent++;
						inject(np);
						global_values.erase(global_values.begin() + i);
					}
					break;
				}
				else
					; // dont break
			}
		}
	}
}

void DmpAgent::set_x()
{
	Tcl & tcl = Tcl::instance();
	char inbuf[128];
	sprintf(inbuf,"$node_(%i) set X_", node_address_);
	tcl.eval(inbuf);
	const char * res = tcl.result();
	char outbuf[128];
	strcpy(outbuf, res);
	x = atof(outbuf);
	//if ( verbose_ ) printf("x = %f\n", x);
}

void DmpAgent::set_y()
{
	Tcl & tcl = Tcl::instance();
	char inbuf[128];
	sprintf(inbuf,"$node_(%i) set Y_", node_address_);
	tcl.eval(inbuf);
	const char * res = tcl.result();
	char outbuf[128];
	strcpy(outbuf, res);
	y = atof(outbuf);
	//if ( verbose_ ) printf("y = %f\n", y);
}

////////////////////////////////////////////////////////////////////////

void DmpAgent::set_range()
{
	double power = transmission_power - margin_ - rx_thr_;
	// power to distance formulae
	transmission_range = pow(10, 0.05 * (power - 20 * log10(4 * acos(-1) / 300000000) - 20 * log10(2.4897 * 1000000000)));
	// TO COMMENT LATER
	//transmission_range = range_max_;
}

void DmpAgent::set_power()
{
	energy_used = energy_used + pow(10, transmission_power/10) * (Scheduler::instance().clock() - transmission_power_change_time) / 1000;
	transmission_power_change_time = Scheduler::instance().clock();
	// distance to power formulae
	double attenuation = 20 * log10(4 * acos(-1) / 300000000) + 20 * log10(2.4897*1000000000) + 20 * log10(transmission_range);
	transmission_power = attenuation + margin_ + rx_thr_;
	if ( transmission_power > tx_max_ )
		transmission_power = tx_max_;
}

void DmpAgent::beacon_timeout()
{
	// do something
	double wait = beacon_period_;
	beacon_timer.resched(wait);
}

void DmpAgent::broadcast_beacon()
{
	if ( verbose_ ) printf("Broadcasting from node %i\n", node_address_);
	set_range();
	set_x();
	set_y();
	dg->update(this);
	std::vector<int>::iterator it = neighbors.begin();
	while ( it != neighbors.end() )
	{
		beacon(*it);
		it++;
	}
}

void DmpAgent::beacon(int daddress)
{
	if ( verbose_ ) printf("Beaconing from %i to %i\n", node_address_, daddress);
	Packet * p = allocpkt();
	hdr_cmn * hc = hdr_cmn::access(p);
	//hc->uid_; // unique id
	//hc->iface_; // receiving interface (label)
	//hc->send_time_ = Scheduler::instance().clock();
	hc->size_ = 128; // simulated packet size in BYTES

	hdr_ip * hi = hdr_ip::access(p);
	hi->src_.addr_ = node_address_;
	hi->dst_.addr_ = daddress;
	hi->src_.port_ = 10;
	hi->dst_.port_ = 10;
	hi->ttl_ = 64;

	hdr_dmp * hd = hdr_dmp::access(p);
	hd->dmp_subtype_ = DMP_BEACON;
	hd->src_address_ = node_address_;
	hd->dst_address_ = daddress;
	hd->timestamp_ = Scheduler::instance().clock();
	hd->x_ = x;
	hd->y_ = y;
	hd->speed_ = 0;
	hd->heading_ = 0;
	hd->comm_nb_ = communicators.size();
	send(p, 0);
}

void DmpAgent::process_beacon(Packet * p)
{
	set_x();
	set_y();
	hdr_dmp * hd = hdr_dmp::access(p);
	double ds = sqrt(pow(hd->x_ - x, 2) + pow(hd->y_ - y, 2));
	communicators[ds] = hd->src_address_;
	while ( static_cast<int>(communicators.size()) > min_connect_ )
	{
		communicators.erase(--communicators.end());
	}
	std::map<double, int>::iterator it = --communicators.end();
	transmission_range = (*it).first;
	//transmission_range = range_max_;
	set_power();
	// NO REPLY FOR THE MOMENT
}

void DmpAgent::process_reply(Packet * rp)
{
	// paquet ne retournera pas si range < range emetteur !!!
	hdr_ip * rhi = hdr_ip::access(rp);
	hdr_dmp * rhd = hdr_dmp::access(rp);
	Packet * p = allocpkt();
	hdr_cmn * hc = hdr_cmn::access(p);
	hc->size_ = 128;
	hdr_ip * hi = hdr_ip::access(p);
	hi->src_.addr_ = node_address_;
	hi->dst_.addr_ = rhi->src_.addr_;
	hi->src_.port_ = 10;
	hi->dst_.port_ = 10;
	hi->ttl_ = 64;
	hdr_dmp * hd = hdr_dmp::access(p);
	hd->dmp_subtype_ = DMP_BEACON_REPLY;
	hd->timestamp_ = Scheduler::instance().clock() - rhd->timestamp_;
	hd->timestamp_ = Scheduler::instance().clock();
	//Packet::free(rp);
	//send(p, 0);

	set_range();
	dg->update(this);
	std::vector<int>::iterator it = neighbors.begin();
	while ( it != neighbors.end() )
	{
		if ( *it == rhi->src_.addr_ )
			// NOOO:unicast(*it);
			send(p, 0);
		else
			if ( verbose_ ) printf("Beacon sender is too far away\n");
		it++;
	}
}

void DmpAgent::init_beacon_timer()
{
	if ( beacon_timer.status() != TIMER_IDLE )
	{
		beacon_timer.cancel();
		if ( verbose_ ) printf("Timer canceled\n");
	}
}

////////////////////////////////////////////////////////////////////////

demigod::demigod()
{
	cc_nb = 0;
	cc_cn = 0;
	mcc_nb = 0;
	mcc_cn = 0;
	net = 0;
	tot_pkts_sent = 0;
	gen_pkts_sent = 0;
	nodes.clear();
}

void demigod::output()
{
	DmpAgent * sk = nodes[0]; // sink
	std::vector<double> total_sent, total_recv, aggr_ratio, energy_ratio, total_sent_fast, total_energy;
	int i;
	for ( i = 1; i < static_cast<int>(nodes.size()); i++ )
	{
		DmpAgent * n = nodes[i];
		total_sent.push_back(n->pkt_sent);
		total_sent_fast.push_back(n->pkt_sent_fast);
		total_recv.push_back(n->pkt_recv);
		if ( n->pkt_recv > 0 )
		{
			aggr_ratio.push_back(1.0 - static_cast<double>(n->pkt_sent) / n->pkt_recv);
		}
		total_energy.push_back(n->node_pt->energy_model()->energy());
		energy_ratio.push_back(n->node_pt->energy_model()->energy() / n->node_pt->energy_model()->initialenergy()); // if initial energy varies with sensor
		//std::cout << "node "<<i<<" has sent "<<n->pkt_sent<<" and rec "<<n->pkt_recv<<"\n";
	}
	if ( sk->e2e_delay.size() < 2 || aggr_ratio.size() < 2 )
	{
		std::cout << "error: no aggr_ratio/e2e_delay data available, skipping this run\n";
		return;
	}
	ostringstream oss;
	oss << sk->seed_ << " " << sk->net_type_ << " " << sk->nodes_nb_ << " " << sk->algorithm_ << " " << sk->message_period_ << " " <<
		//dmp::sample_sum(total_sent) / dmp::sample_sum(total_recv) << " "<<
		//dmp::sample_standard_deviation(total_sent) / dmp::sample_standard_deviation(total_recv) << " " <<
		///tot_pkts_sent << " " << gen_pkts_sent << " " << dmp::sample_sum(total_sent_fast) << " " << sk->e2e_delay.size() << " " <<
		dmp::sample_mean(aggr_ratio) << " "<< dmp::sample_standard_deviation(aggr_ratio) << " " <<
		dmp::sample_mean(sk->e2e_delay) << " " << dmp::sample_standard_deviation(sk->e2e_delay) << " " <<
		///dmp::sample_sum(total_energy) << " " << nodes.size() * sk->node_pt->energy_model()->initialenergy() << " " <<
		dmp::sample_mean(energy_ratio) << " "<< dmp::sample_standard_deviation(energy_ratio) << " " << "\n";
	//std::cout << "seed net size algo period aggr aggr_sd delay delay_sd energy energy_sd\n";
	//std::cout << oss.str();
	std::ostringstream fn;
	fn << "data-dmp-" << sk->net_type_ << "-" << sk->nodes_nb_ << "-" << sk->algorithm_ << "-" << sk->message_period_ << ".txt" << '\0';
	std::ofstream f;
	f.open(fn.str().c_str(), ios::app);
	f << oss.str();
	//f << "cc=" << cc_nb / cc_cn << "\n";
	//f << "mcc=" << mcc_nb / mcc_cn << "\n";
	f.close();
}

void demigod::init(DmpAgent * a)
{
	nodes[a->node_address_] = a;
	if ( a->node_address_ == 0 )
	{
		// init MT PRNG
		dmp::init_genrand(static_cast<unsigned long>(a->seed_));
		return; // don't use matrices
		distance_matrix.clear();
		connect_matrix.clear();
		int i, j;
		for ( i = 0; i < a->nodes_nb_; i++ )
		{
			distance_vector dv;
			dv.clear();
			connect_vector cv;
			cv.clear();
			for ( j = 0; j < a->nodes_nb_; j++ )
			{
				double r = 0;
				dv.push_back(r);
				int e = 0;
				cv.push_back(e);
			}
			distance_matrix.push_back(dv);
			connect_matrix.push_back(cv);
		}
	}
}

// WORKS IF INDEX = NODE ADDRESS
void demigod::update(DmpAgent * a)
{
	return;
	a->set_x();
	a->set_y();
	int i;
	for ( i = 0; i < a->nodes_nb_; i++ )
	{
		DmpAgent * n = nodes[i];
		if ( n != a )
		{
			n->set_x();
			n->set_y();
		}
		distance_matrix[a->node_address_][i] = sqrt(pow(n->x - a->x, 2) + pow(n->y - a->y, 2));
	}
	a->neighbors.clear();
	a->maxneighbors.clear();
	a->set_range();
	for ( i = 0; i < a->nodes_nb_; i++ )
	{
		DmpAgent * n = nodes[i];
		double dm = distance_matrix[a->node_address_][i];
		if ( dm < 0 )
			printf("err: negative distance\n");
		else if ( dm <= a->transmission_range )
		{
			connect_matrix[a->node_address_][i] = 1;
			if ( n != a )
			{
				a->neighbors.push_back(i);
				a->maxneighbors.push_back(i);
			}
		}
		else if ( dm <= a->range_max_ )
		{
			connect_matrix[a->node_address_][i] = 2;
			if ( n != a )
			{
				a->maxneighbors.push_back(i);
			}
		}
		else
			connect_matrix[a->node_address_][i] = 0;
	}
	net = 2;
	mcc_nb += tarjan();
	mcc_cn++;
	net = 1;
	cc_nb += tarjan();
	cc_cn++;
	output();
}

int demigod::tarjan()
{
	int cpt = 0, m = 0; // m = nb of scc already found
	std::stack<DmpAgent *> P;
	int i;
	for ( i = 0; i < static_cast<int>(nodes.size()); i++ )
		nodes[i]->prefix = 0;
	for ( i = 0; i < static_cast<int>(nodes.size()); i++ )
		if ( nodes[i]->prefix == 0 )
			strongly_connected_component(nodes[i], &cpt, &m, &P);
	return m;
}

void demigod::strongly_connected_component(DmpAgent * x, int * cpt, int * m, std::stack<DmpAgent *> * P)
{
	int j;
	DmpAgent * y;
	(*cpt)++;
	x->prefix = (*cpt); // prefix numbering of x
	int min = (*cpt); // x->back <= x->prefix, min = temporary value of x->back
	P->push(x);
	int nsize;
	if ( net == 1 )
		nsize = static_cast<int>(x->neighbors.size());
	else if ( net == 2 )
		nsize = static_cast<int>(x->maxneighbors.size());
	else
		printf("err: wrong net value\n");
	for ( j = 0; j < nsize; j++ )
	{
		if ( net == 1 )
			y = nodes[x->neighbors[j]];
		else if ( net == 2 )
			y = nodes[x->maxneighbors[j]];
		else
			printf("err: wrong net value\n");
		if ( y->prefix == 0 ) // first encounter of y
		{
			strongly_connected_component(y, cpt, m, P);
			if ( y->back < min )
				min = y->back;
		}
		else // y already encountered, x->y is a back_edge or cross_edge
		{
			if ( y->prefix < min ) // if y->prefix = n+1 then y has already been put in another scc
				min = y->prefix;
		}
	}
	x->back = min;
	if ( x->back == x->prefix ) // we get a new scc
	{
		(*m)++;
		while ( P->top() != x )
		{
			y = P->top();
			y->cc_no = (*m);
			y->prefix = nodes.size() + 1;
			P->pop();
		}
		x->prefix = nodes.size() + 1;
		x->cc_no = (*m);
		P->pop();
	}
}

////////////////////////////////////////////////////////////////////////

namespace dmp
{
	double dist(double x1, double y1, double x2, double y2)
	{
		return sqrt(pow(x1 - x2, 2) + pow(y1 - y2, 2));
	}

	// return the nearest int of d
	int round(double d)
	{
		double n = 0;
		if ( fabs(modf(d, &n)) <= 0.5 )
			return static_cast<int>(n);
		else if ( d > 0 )
			return static_cast<int>(n) + 1;
		else
			return static_cast<int>(n) - 1;
	}

	// proba must be expressed in per cent (e.g. 37.4 %)
	bool roll(double pc_probability)
	{
		if ( random(1, 10000) <= round(100 * pc_probability) )
			return true;
		else
			return false;
	}

	int random(int inf, int sup)
	{
		// Classic ANSI C rand() LC-PRNG
		// put next instruction at the beginning of the program
		// srand((unsigned)time(0));
		// return round((sup - inf) * static_cast<double>(rand()) / RAND_MAX) + inf;

		// Mersenne Twister GFSR-PRNG
		// put next instruction at the beginning of the program
		// init_genrand((unsigned long)s);
		return round((sup - inf) * static_cast<double>(genrand_int32()) / DMP_INT32_MAX) + inf;
	}

	double rzto()
	{
		return static_cast<double>(dmp::random(0, 1000000)) / 1000000.0;
	}

	std::vector<double> extract_vec(std::map<double, double> & mm, bool key)
	{
		std::vector<double> v;
		std::map<double, double>::iterator mmit = mm.begin();
		while ( mmit != mm.end() )
		{
			if ( key )
				v.push_back(mmit->first);
			else
				v.push_back(mmit->second);
			mmit++;
		}
		return v;
	}

	double sample_sum(std::vector<double> & sample_data)
	{
		int i;
		double sum = 0;
		for ( i = 0; i < sample_data.size(); i++ )
			sum += sample_data[i];
		return sum;
	}

	double sample_min(std::vector<double> & sample_data)
	{
		int i;
		double min = dmp::dbl_max;
		for ( i = 0; i < sample_data.size(); i++ )
			if ( min > sample_data[i] )
				min = sample_data[i];
		return min;
	}

	double sample_max(std::vector<double> & sample_data)
	{
		int i;
		double max = -1.0 * dmp::dbl_max;
		for ( i = 0; i < sample_data.size(); i++ )
			if ( max < sample_data[i] )
				max = sample_data[i];
		return max;
	}

	double sample_perc(double perc, std::vector<double> & sample_data)
	{
		std::cout << "cant use std::sort\n";
		return 0;
		//std::sort(sample_data.begin(), sample_data.end()); // change order to ascending with operator <
		perc = perc / 100.0;
		int i = 0;
		if ( perc < 0 )
			i = 0;
		else if ( round(perc * sample_data.size()) >= sample_data.size() )
			i = sample_data.size() - 1;
		else
			i = round(perc * sample_data.size());
		return sample_data[i];
	}

	double sample_mean(std::vector<double> & sample_data)
	{
		return sample_sum(sample_data) / sample_data.size();
	}

	double sample_standard_deviation(std::vector<double> & sample_data)
	{
		int n = sample_data.size(), i;
		double s = 0, m = sample_mean(sample_data);
		for ( i = 0; i < n; i++ )
			s += pow(sample_data[i] - m, 2);
		return sqrt(s / (n - 1));
	}

	double sample_standard_error_of_the_mean(std::vector<double> & sample_data)
	{
		return sample_standard_deviation(sample_data) / sqrt(static_cast<double>(sample_data.size()));
	}

	double sample_relative_statistical_error(double confidence_level, std::vector<double> & sample_data)
	{
		double ccl = 1 - confidence_level;
		int n = sample_data.size();
		double sm = sample_mean(sample_data);
		if ( sm )
			return AStudT(ccl, n - 1) * sample_standard_error_of_the_mean(sample_data) / sm;
		else
		{
			if ( !sample_standard_error_of_the_mean(sample_data) )
				return 0;
			else
			{
				std::cout << "rse div by 0\n";
				return 1; // infinite rse
			}
		}
	}

	// John C. Pezzullo at johnp71@aol.com
	double StatCom(double q, int i, int j, int b)
	{
		double zz = 1, z = zz;
		int k = i;
		while ( k <= j )
		{
			zz = zz * q * k / (k - b);
			z = z + zz;
			k = k + 2;
		}
		return z;
	}

	double StudT(double t, int n)
	{
		t = fabs(t);
		double w = t / sqrt(static_cast<double>(n));
		double th = atan(w);
		if ( n == 1 )
		{
			return 1 - th / (dmp::pi / 2);
		}
		double sth = sin(th);
		double cth = cos(th);
		if( (n % 2) == 1 )
			return 1 - (th + sth * cth * StatCom(cth * cth, 2, n - 3, -1) ) / (dmp::pi / 2);
		else
			return 1 - sth * StatCom(cth * cth, 1, n - 3, -1);
	}

	double AStudT(double p, int n)
	{
		double v = 0.5, dv = 0.5, t = 0;
		while( dv > 1e-6 )
		{
			t = 1 / v - 1;
			dv = dv / 2;
			if ( StudT(t, n) > p )
				v = v - dv;
			else
				v = v + dv;
		}
		return t;
	}

	/*
		   A C-program for MT19937, with initialization improved 2002/1/26.
		   Coded by Takuji Nishimura and Makoto Matsumoto.

		   Before using, initialize the state by using init_genrand(seed)
		   or init_by_array(init_key, key_length).

		   Copyright (C) 1997 - 2002, Makoto Matsumoto and Takuji Nishimura,
		   All rights reserved.

		   Redistribution and use in source and binary forms, with or without
		   modification, are permitted provided that the following conditions
		   are met:

			 1. Redistributions of source code must retain the above copyright
				notice, this list of conditions and the following disclaimer.

			 2. Redistributions in binary form must reproduce the above copyright
				notice, this list of conditions and the following disclaimer in the
				documentation and/or other materials provided with the distribution.

			 3. The names of its contributors may not be used to endorse or promote
				products derived from this software without specific prior written
				permission.

		   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
		   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
		   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
		   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
		   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
		   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
		   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
		   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
		   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
		   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
		   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

		   Any feedback is very welcome.
		   http://www.math.keio.ac.jp/matumoto/emt.html
		   email: matumoto@math.keio.ac.jp
	 */

	//	#include <stdio.h>

	/* Period parameters */
	#define N 624
	#define M 397
	#define MATRIX_A 0x9908b0dfUL   /* constant vector a */
	#define UPPER_MASK 0x80000000UL /* most significant w-r bits */
	#define LOWER_MASK 0x7fffffffUL /* least significant r bits */

	static unsigned long mt[N]; /* the array for the state vector  */
	static int mti=N+1; /* mti==N+1 means mt[N] is not initialized */

	/* initializes mt[N] with a seed */
	void init_genrand(unsigned long s)
	{
		mt[0]= s & 0xffffffffUL;
		for (mti=1; mti<N; mti++) {
			mt[mti] =
					(1812433253UL * (mt[mti-1] ^ (mt[mti-1] >> 30)) + mti);
			/* See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier. */
			/* In the previous versions, MSBs of the seed affect   */
			/* only MSBs of the array mt[].                        */
			/* 2002/01/09 modified by Makoto Matsumoto             */
			mt[mti] &= 0xffffffffUL;
			/* for >32 bit machines */
		}
	}

	/* initialize by an array with array-length */
	/* init_key is the array for initializing keys */
	/* key_length is its length */
	void init_by_array(unsigned long init_key[], unsigned long key_length)
	{
		int i, j, k;
		init_genrand(19650218UL);
		i=1; j=0;
		k = (N>key_length ? N : key_length);
		for (; k; k--) {
			mt[i] = (mt[i] ^ ((mt[i-1] ^ (mt[i-1] >> 30)) * 1664525UL))
						  + init_key[j] + j; /* non linear */
			mt[i] &= 0xffffffffUL; /* for WORDSIZE > 32 machines */
			i++; j++;
			if (i>=N) { mt[0] = mt[N-1]; i=1; }
			if (j>=(int)key_length) j=0;
		}
		for (k=N-1; k; k--) {
			mt[i] = (mt[i] ^ ((mt[i-1] ^ (mt[i-1] >> 30)) * 1566083941UL))
						  - i; /* non linear */
			mt[i] &= 0xffffffffUL; /* for WORDSIZE > 32 machines */
			i++;
			if (i>=N) { mt[0] = mt[N-1]; i=1; }
		}

		mt[0] = 0x80000000UL; /* MSB is 1; assuring non-zero initial array */
	}

	/* generates a random number on [0,0xffffffff]-interval */
	unsigned long genrand_int32(void)
	{
		unsigned long y;
		static unsigned long mag01[2]={0x0UL, MATRIX_A};
		/* mag01[x] = x * MATRIX_A  for x=0,1 */

		if (mti >= N) { /* generate N words at one time */
			int kk;

			if (mti == N+1)   /* if init_genrand() has not been called, */
				init_genrand(5489UL); /* a default initial seed is used */

			for (kk=0;kk<N-M;kk++) {
				y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
				mt[kk] = mt[kk+M] ^ (y >> 1) ^ mag01[y & 0x1UL];
			}
			for (;kk<N-1;kk++) {
				y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
				mt[kk] = mt[kk+(M-N)] ^ (y >> 1) ^ mag01[y & 0x1UL];
			}
			y = (mt[N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK);
			mt[N-1] = mt[M-1] ^ (y >> 1) ^ mag01[y & 0x1UL];

			mti = 0;
		}

		y = mt[mti++];

		/* Tempering */
		y ^= (y >> 11);
		y ^= (y << 7) & 0x9d2c5680UL;
		y ^= (y << 15) & 0xefc60000UL;
		y ^= (y >> 18);

		return y;
	}

	/* generates a random number on [0,0x7fffffff]-interval */
	long genrand_int31(void)
	{
		return (long)(genrand_int32()>>1);
	}

	/* generates a random number on [0,1]-double-interval */
	double genrand_double1(void)
	{
		return genrand_int32()*(1.0/4294967295.0);
		/* divided by 2^32-1 */
	}

	/* generates a random number on [0,1)-double-interval */
	double genrand_double2(void)
	{
		return genrand_int32()*(1.0/4294967296.0);
		/* divided by 2^32 */
	}

	/* generates a random number on (0,1)-double-interval */
	double genrand_double3(void)
	{
		return (((double)genrand_int32()) + 0.5)*(1.0/4294967296.0);
		/* divided by 2^32 */
	}

	/* generates a random number on [0,1) with 53-bit resolution*/
	double genrand_res53(void)
	{
		unsigned long a=genrand_int32()>>5, b=genrand_int32()>>6;
		return(a*67108864.0+b)*(1.0/9007199254740992.0);
	}
	/* These real versions are due to Isaku Wada, 2002/01/09 added */
}

// EOF
