# -*-	Mode:tcl; tcl-indent-level:4; tab-width:4; indent-tabs-mode:t -*-

#
# Copyright (c) 2014, Damien Magoni.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#    This product includes software developed by Damien Magoni.
# 4. The name of Damien Magoni may not be used
#    to endorse or promote products derived from this software without
#    specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY DAMIEN MAGONI ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL DAMIEN MAGONI OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#

# DMP for Partial Data Aggregation

set pyloop 0

if {$pyloop == 1} {
	set iv(seed)                 [lindex $argv 0]
	set iv(sim_duration)         [lindex $argv 1]
	set iv(net_type)             [lindex $argv 2]
	set iv(net_size)             [lindex $argv 3]
	set iv(algorithm)            [lindex $argv 4]
	set iv(message_period)       [lindex $argv 5]
	set iv(local_wait_period)    $iv(message_period) ;#[lindex $argv 6]
	set iv(global_wait_period)   $iv(message_period) ;#[lindex $argv 7]
	} else {
	set iv(seed)                 0 ;#[lindex $argv 0]
	set iv(net_type)             4 ;#[lindex $argv 1]
	set iv(net_size)           100 ;#[lindex $argv 2]
	set iv(algorithm)            4 ;#[lindex $argv 3]
	set iv(sim_duration)       600 ;#[lindex $argv 4]
	set iv(message_period)      60 ;#[lindex $argv 5]
	set iv(local_wait_period)    6 ;#[lindex $argv 6]
	set iv(global_wait_period)  12 ;#[lindex $argv 7]
		
	set iv(beacon_period)        0 ;#[lindex $argv 6]
	set iv(min_connect)          0 ;#[lindex $argv 7]
	set iv(urgent_pc)            0 ;#[lindex $argv 8]
}

set val(nam)             1
set val(verbose)         1
set val(trace)           ON
set val(nid)             10.0  ;# node inter distance
set val(x)               100.0 ;# X dimension of the topography
set val(y)               100.0 ;# Y dimension of the topography
set val(regps)	         64    ;# max value is 100 as header is 27 and packet <= 127
set val(urgps)	         [expr $val(regps) - 1]
set val(tx_max)          20.0
set val(rx_thr)          -70.0
set val(margin)          10.0
set val(range_max)       95.9

set val(chan)       Channel/WirelessChannel
#set val(prop)       Propagation/FreeSpace ;# valid for short distances
set val(prop)       Propagation/TwoRayGround
set val(ant)        Antenna/OmniAntenna
set val(netif)      Phy/WirelessPhy/802_15_4
set val(mac)        Mac/802_15_4
set val(ifq)        Queue/DropTail/PriQueue
set val(ifqlen)     100    ;# max packet in ifq
set val(ll)         LL
set val(arp)        AODV

set ns		    [new Simulator]
set topo	    [new Topography]
set tracefd	    [open trace-dmp.tr w]
set namtracefd      [open anim-dmp.nam w]

$ns use-newtrace
$ns trace-all $tracefd
if {$val(nam) == 1} {
    $ns namtrace-all-wireless $namtracefd $val(x) $val(y)
    $ns puts-nam-traceall {# nam4wpan #} ;# inform nam that this is a trace file for wpan (special handling needed)
}

#-----------------------------------------------------------
#Tcl command: Mac/802_15_4 wpanCmd verbose [on/off]
#Tcl command: Mac/802_15_4 wpanCmd txOption [2/4/0]
#Tcl command: Mac/802_15_4 wpanCmd ack4data [on/off]
#Tcl command: Mac/802_15_4 wpanCmd callBack [0/1/2]
#Tcl command: Mac/802_15_4 wpanCmd link-down <src> <dst>
#Tcl command: Mac/802_15_4 wpanCmd link-up <src> <dst>
#-----------------------------------------------------------
Mac/802_15_4 wpanCmd verbose off                    ;#Run simulation in verbose mode or non-verbose mode;
#Mac/802_15_4 wpanCmd backoffAlgo $iv_algorithm      ;# Choose the proper algo to evaluate for EBSA

#-----------------------------------------------------------------
#Tcl command: Mac/802_15_4 wpanNam namStatus [on/off]
#Tcl command: Mac/802_15_4 wpanNam emHandling [on/off]
#Tcl command: Mac/802_15_4 wpanNam PANCoorClr [clrName]
#Tcl command: Mac/802_15_4 wpanNam CoorClr [clrName]
#Tcl command: Mac/802_15_4 wpanNam DevClr [clrName]
#Tcl command: Mac/802_15_4 wpanNam ColFlashClr [clrName]
#Tcl command: Mac/802_15_4 wpanNam NodeFailClr [clrName]
#Tcl command: Mac/802_15_4 wpanNam PlaybackRate [step_in_ms]
#Tcl command: Mac/802_15_4 wpanNam FlowClr [-p <packet_type_name>] [-s <src>] [-d <dst>] [-c <clrName>]
#-----------------------------------------------------------------
Mac/802_15_4 wpanNam namStatus on		    ;# default = off (should be turned on before other 'wpanNam' commands can work)
#Mac/802_15_4 wpanNam DevClr black	            
Mac/802_15_4 wpanNam ColFlashClr gold	            ;# default = gold
Mac/802_15_4 wpanNam NodeFailClr grey 	            ;# default = grey
Mac/802_15_4 wpanNam PlaybackRate 50.0ms
Mac/802_15_4 wpanNam FlowClr -p AODV -c blue
Mac/802_15_4 wpanNam FlowClr -p ARP  -c green
Mac/802_15_4 wpanNam FlowClr -p MAC  -c cyan
Mac/802_15_4 wpanNam FlowClr -p DMP  -c red

#Mac/802_15_4 wpanNam FlowClr -p tcp -s 19 -d 6 -c blue
#Mac/802_15_4 wpanNam FlowClr -p ack -s 6 -d 19 -c navy
#Mac/802_15_4 wpanNam FlowClr -p cbr -s 0 -d 1 -c yellow
#Mac/802_15_4 wpanNam FlowClr -p exp -s 0 -d 1 -c brown

# A C program is provided at ~ns/indep-utils/propagation/threshold.cc to compute the receiving threshold
# For model 'TwoRayGround'
set dist(5m)  7.69113e-06
set dist(10m) 1.92278e-06
set dist(11m) 1.58908e-06
set dist(12m) 1.33527e-06
set dist(13m) 1.13774e-06
set dist(14m) 9.81011e-07
set dist(15m) 8.54570e-07
set dist(20m) 4.80696e-07
set dist(25m) 3.07645e-07
set dist(30m) 2.13643e-07
set dist(35m) 1.56962e-07
set dist(40m) 1.20174e-07
set dist(50m) 7.69113e-08
# For model 'FreeSpace', Fr = 2.4GHz, Pt = 1e-3
set dist(4.0m) 5.93432e-09
set dist(4.5m) 4.68884e-09
set dist(6.0m) 2.63747e-09
set dist(9.0m) 1.17221e-09

Phy/WirelessPhy set CSThresh_ $dist(25m)  ;# Carrier Sense Threshold, should we put RTmin?
Phy/WirelessPhy set RXThresh_ $dist(12m)  ;# Receiver Threshold: RTmin <= -85dBm = 3,1622776602e-12,  RXTmax >= -20dBm = 1e-5
Phy/WirelessPhy set CPThresh_ 10.0	  ;# Collision Threshold
#When multi-frames are received simultaneously by one mobile node,
#it calculates the ratio of the strongest frame’s signal strength to the signal strength sum of other frames.
#If it is larger than CPThresh_, the frame will be received correctly and other frames are ignored.
#Otherwise, all frames are collided and discarded.
Phy/WirelessPhy set bandwidth_ 250000      ;# (bit/s)
#Phy/WirelessPhy set Rb_ 2e6
#Phy/WirelessPhy set Pt_ 0.001             ;# Transmitter Signal Power (Watt) >= -3dBm = 5,0118723363e-4
#Phy/WirelessPhy set freq_ 2450e+6         ;# Frequency (Hz)
#Phy/WirelessPhy set L_ 1.0

$topo load_flatgrid $val(x) $val(y)
create-god [expr $iv(net_size)]

set chan_1_ [new $val(chan)]
$ns node-config \
    -adhocRouting $val(arp) \
    -llType $val(ll) \
    -macType $val(mac) \
    -ifqType $val(ifq) \
    -ifqLen $val(ifqlen) \
    -antType $val(ant) \
    -propType $val(prop) \
    -phyType $val(netif) \
    -topoInstance $topo \
    -agentTrace ON \
    -routerTrace ON \
    -macTrace ON \
    -movementTrace OFF \
    -channel $chan_1_ \
    -wiredRouting OFF \
    -mobileIP OFF \
    -energyModel "EnergyModel" \
    -initialEnergy 1000 \
    -rxPower 35.28e-3 \
    -txPower 31.32e-3 \
    -idlePower 712e-6 \
    -sleepPower 144e-9

#-ifqLen 2 \ ;# for all nodes except gw/sink

global defaultRNG
$defaultRNG seed $iv(seed)                   ;# should remain fixed to 1? no
set message_startRNG [new RNG]
for {set j 0} {$j < $iv(seed)} {incr j} {    ;# to avoid stream overlaps
    $message_startRNG next-substream
}

set ms [new RandomVariable/Uniform]
$ms use-rng $message_startRNG
$ms set min_ 0
$ms set max_ 1000000

set posrndzr [new RandomVariable/Uniform]
$ms use-rng $message_startRNG
$ms set min_ 0
$ms set max_ 100

for {set i 0} {$i < $iv(net_size)} {incr i} {
    set node_($i) [$ns node]
    $node_($i) random-motion 0
    $node_($i) NodeClr Black
    #$node_($i) radius 12 ;# not used
}

for {set i 0} {$i < $iv(net_size)} {incr i} {
    set dmp_($i) [new Agent/DMP]
    $node_($i) attach $dmp_($i) 10 ;# port number of the agent
    $dmp_($i) set node_handle_ $node_($i)
    $dmp_($i) set node_id_ [$node_($i) id]
    $dmp_($i) set node_address_ [$node_($i) node-addr]
    $dmp_($i) set mac_handle_ [$node_($i) set mac_(0)]
    $dmp_($i) set seed_ $iv(seed)
    $dmp_($i) set verbose_ $val(verbose)
    $dmp_($i) set net_type_ $iv(net_type)
    $dmp_($i) set nodes_nb_ $iv(net_size)
    $dmp_($i) set message_period_ $iv(message_period)
    $dmp_($i) set local_wait_period_ $iv(local_wait_period)
    $dmp_($i) set global_wait_period_ $iv(global_wait_period)
    $dmp_($i) set algorithm_ $iv(algorithm)
}

# Define topology
if {$iv(net_type) == 0} {
    # test
    for {set i 0} {$i < $iv(net_size) } {incr i} {
	$node_($i) set X_ [expr 280.0 - $i * $val(nid)] 
	$node_($i) set Y_ 20.0
	$node_($i) set Z_ 0.0
    }
} elseif {$iv(net_type) == 1} {
    # highway, highway profile: 3.5m between each SN
    set h [expr $iv(net_size) / 10]    ;#round(sqrt($iv(net_size)))]
    set gmyX [expr 20.0 + $val(nid) * double($h) * 6.0 / 2.0] ;# gw in the external middle
    set gmyY [expr 20.0 - $val(nid)]
    #puts "gw x = $gmyX y = $gmyY"
    $gw set X_ $gmyX
    $gw set Y_ $gmyY
    $gw set Z_ 0.0
    for {set i 0} {$i < $iv(net_size) } {incr i} {
	if { [expr ($i % 10) / 5] == 0 } {
	    set myX [expr 20.0 + ($i / 10) * 6 * $val(nid)]
	    set myY [expr 20.0 + $val(nid) * ($i % 10)]
	} else {
	    set myX [expr 20.0 + $val(nid) + ($i / 10) * 6 * $val(nid) + (($i % 10) - 5) * $val(nid)]
	    set myY [expr 20.0 + 0]
	}
	#puts "node $i x = $myX y = $myY"
	$node_($i) set X_ $myX
	$node_($i) set Y_ $myY
	$node_($i) set Z_ 0.0
    }
} elseif {$iv(net_type) == 2} {
    # traffic cross
    set h [expr round(sqrt($iv(net_size)))]
    set gmyX [expr 20.0 + $val(nid) * double($h)]
    set gmyY [expr 20.0 + $val(nid) * double($h)]
    #puts "gw x = $gmyX y = $gmyY"
    $gw set X_ $gmyX
    $gw set Y_ $gmyY
    $gw set Z_ 0.0
    set mypi 3.141592653589793 ;# 23846 26433 83279 50288
    set step [expr 2.0 * $mypi / double($h)]
    for {set i 0} {$i < $iv(net_size) } {incr i} {
	set module [expr $val(nid) * double($i / $h + 1)]
	set myX [expr $gmyX + $module * cos(double($i % $h) * $step)]
	set myY [expr $gmyY + $module * sin(double($i % $h) * $step)]
	#puts "node $i x = $myX y = $myY"
	$node_($i) set X_ $myX
	$node_($i) set Y_ $myY
	$node_($i) set Z_ 0.0
    }
} elseif {$iv(net_type) == 3} {
    # traffic circle
    set h [expr round(sqrt($iv(net_size)))]
    set gmyX [expr 20.0 + $val(nid) * double($h)]
    set gmyY [expr 20.0 + $val(nid) * double($h)]
    #puts "gw x = $gmyX y = $gmyY"
    $gw set X_ $gmyX
    $gw set Y_ $gmyY
    $gw set Z_ 0.0
    set mypi 3.141592653589793 ;# 23846 26433 83279 50288
    set step [expr 2.0 * $mypi / double($h)]
    for {set i 0} {$i < $iv(net_size) } {incr i} {
	set module [expr $val(nid) * double($i / $h + 1)]
	set myX [expr $gmyX + $module * cos(double($i % $h) * $step)]
	set myY [expr $gmyY + $module * sin(double($i % $h) * $step)]
	#puts "node $i x = $myX y = $myY"
	$node_($i) set X_ $myX
	$node_($i) set Y_ $myY
	$node_($i) set Z_ 0.0
    }
} elseif {$iv(net_type) == 5} {
    # should be changed to create T shape!
    set h [expr round(ceil($iv(net_size) / 2.0))] ;# should round up
    set gmyX [expr 20.0 + $val(nid) * double($h) / 2.0]
    set gmyY [expr 20.0 + $val(nid) * 0.0]
    #puts "gw x = $gmyX y = $gmyY"
    $gw set X_ $gmyX
    $gw set Y_ $gmyY
    $gw set Z_ 0.0
    for {set i 0} {$i < $iv(net_size) } {incr i} {
	set myX [expr 20.0 + $val(nid) * double($i % $h)]
	set myY [expr 20.0 + $val(nid) * double($i / $h)]
	#puts "node $i x = $myX y = $myY"
	$node_($i) set X_ $myX
	$node_($i) set Y_ $myY
	$node_($i) set Z_ 0.0
    }
} elseif {$iv(net_type) == 4} {
    puts "Using a grid topology"
    set h [expr round(sqrt($iv(net_size) / 1.0))] ;# should round up
    #set gmyX [expr 20.0 + $val(nid) * double($h) / 2.0]
    #set gmyY [expr 20.0 + $val(nid) * 0.0]
    #puts "gw x = $gmyX y = $gmyY"
    $node_(0) set X_ 0.0 ;#$gmyX
    $node_(0) set Y_ 0.0 ;#$gmyY
    $node_(0) set Z_ 0.0
    for {set i 1} {$i < $iv(net_size) } {incr i} {
	set myX [expr 0.0 + $val(nid) * double($i % $h)]
	set myY [expr 0.0 + $val(nid) * double($i / $h)]
	#puts "node $i x = $myX y = $myY"
	$node_($i) set X_ $myX
	$node_($i) set Y_ $myY
	$node_($i) set Z_ 0.0
    }
} elseif {$iv(net_type) == 6} {
    # random
    #set h [expr round(ceil($iv(net_size) / 10.0))] ;# should round up
    #set gmyX [expr 20.0 + $val(nid) * double($h) / 2.0]
    #set gmyY [expr 20.0 + $val(nid) * 0.0]
    #puts "gw x = $gmyX y = $gmyY"
    $node_(0) set X_ 0.0
    $node_(0) set Y_ 0.0
    $node_(0) set Z_ 0.0
    for {set i 1} {$i < $iv(net_size) } {incr i} {
        set myX [expr [$posrndzr value] + 0.0]
        set myY [expr [$posrndzr value] + 0.0]
        #puts "node $i x = $myX y = $myY"
        $node_($i) set X_ $myX
        $node_($i) set Y_ $myY
        $node_($i) set Z_ 0.0
    }
} else {
    puts "Unknown network type"
}

if {$val(nam) == 1} {
    for {set i 0} {$i < $iv(net_size)} {incr i} {
	# 20 defines the node size in nam, adjust it to your scenario
	# The function must be called after mobility model is defined
	$ns initial_node_pos $node_($i) 2
    }
}

# Tell nodes when the simulation starts
for {set i 0} {$i < $iv(net_size) } {incr i} {
    $ns at 0.0 "$dmp_($i) dmpstart"
}

$ns at [expr $iv(sim_duration) + 0.1] "$dmp_(0) dmpstat"

# Tell nodes when the simulation ends
for {set i 0} {$i < $iv(net_size) } {incr i} {
    $ns at [expr $iv(sim_duration) + 0.11] "$node_($i) reset"
}

proc finish {} {
    #puts "End of sim"
    global ns tracefd namtracefd
    $ns flush-trace
    close $tracefd
    close $namtracefd
}

if {$val(nam) == 1} {
    $ns at [expr $iv(sim_duration)+0.111] "$ns nam-end-wireless [expr $iv(sim_duration)+0.111]"
}
$ns at [expr $iv(sim_duration)+0.1111] "finish; $ns halt"

#puts $tracefd "M 0.0 nn $iv(net_size) x $val(x) y $val(y) arp $val(arp)"
#puts $tracefd "M 0.0 mob $val(mob) traf $val(traf) seed $seed"

$ns run

# end of script
